# Load bio2 format to Document object
# Note: the tokenizer must use spacy pipeline
# file format: columns split by whitespace
# need to pass the indexes for text and ner columns


def load_doc(data_file):
    with open(data_file, 'r') as f:
        sents = []
        sent_text = []
        ner_tags = []
        sent_ner_tags = []
        for i, line in enumerate(f):
            if line == '\n' and len(sent_text) != 0:
                sents.append(sent_text)
                sent_text = []
                ner_tags.append(sent_ner_tags)
                sent_ner_tags = []
            cols = line.split()
            if len(cols) < 2:
                continue
            sent_text.append(cols[0])
            sent_ner_tags.append(cols[1].strip())

        # doc_text = concatenate_sents(sents)
        # print(doc_text)
        return sents, ner_tags


def concatenate_sents(sentences):
    # sentences is a list of lists of tokens
    sentences = [' '.join(sent) for sent in sentences]
    doc = '\n'.join(sentences)
    return doc

def concatenate_sents_to_one(sentences):
    # sentences is a list of lists of tokens
    sentences = [' '.join(sent) for sent in sentences]
    doc = ' '.join(sentences)
    return doc



from cdeid.display.html_generator import add_html_tags_of_entities, generate_html
import logging

from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


class Token:
    def __init__(self, text, ner_tag):
        self.text = text
        self.ner_tag = ner_tag


class Sentence:
    def __init__(self):
        self.id = 0
        self.text = ''
        self.sentence_number = 0
        self._tokens = []
        self._entities = []  # entity (start_idx, end_idx, type)
        self._ents = []  # entity (text, type)
        self._html_text = ''

    @property
    def tokens(self):
        return self._tokens

    @tokens.setter
    def tokens(self, tokens):
        self._tokens = tokens

    def add_token(self, token):
        self._tokens.append(token)

    # entity (start_idx, end_idx, type)
    @property
    def entities(self):
        self._entities = self._parse_entities('index')
        return self._entities

    # entity (text, type)
    @property
    def ents(self):
        self._ents = self._parse_entities('plain')
        return self._ents

    # @property
    # def html_text(self):
    #     return self._html_text
    #
    # @html_text.setter
    # def html_text(self, html_text):
    #     self._html_text = html_text

    # @property
    # def annotated_text(self):
    #     tmp_text = self.text
    #     offset = 0
    #     for entity in self._entities:
    #         tmp_text = tmp_text[:entity[0] + offset] + '<**' + entity[2] + '**>' + tmp_text[entity[1] + offset:]
    #         offset += (len(entity[2]) + 6 - (entity[1] - entity[0]))
    #     return tmp_text

    # mode: index (start_idx, end_idx, type) , plain (text, type)
    def _parse_entities(self, mode='index'):
        entity_list = []

        current_index = 0
        entity_flag = False
        entity_type = None
        entity_tokens = []
        for i, token in enumerate(self._tokens):
            if token.ner_tag.startswith('B-'):
                entity_type = token.ner_tag[2:].strip()
                entity_flag = True
                if not (token.text.isspace() or token.text == ''):
                    entity_tokens.append(token)
            if token.ner_tag.startswith('I-') and entity_flag:
                entity_tokens.append(token)
            if (token.ner_tag == 'O' or i == len(self._tokens) - 1) and entity_flag:
                entity_flag = False

                # check if the last token is whitespace
                if len(entity_tokens) != 0 and (entity_tokens[-1].text.isspace() or entity_tokens[-1].text == ''):
                    entity_tokens.pop()
                if len(entity_tokens) == 0:
                    entity_tokens = []
                    entity_type = None
                    continue

                start_idx = self.text.index(entity_tokens[0].text, current_index)
                end_idx = self.text.index(entity_tokens[-1].text, current_index) + len(entity_tokens[-1].text)
                if mode == 'index':
                    entity_list.append((start_idx, end_idx, entity_type))
                elif mode == 'plain':
                    entity_list.append((self.text[start_idx:end_idx], entity_type))
                else:
                    raise Exception('Wrong entity mode: index or plain')
                entity_tokens = []
                entity_type = None
                current_index = end_idx

        return entity_list


class Document:
    def __init__(self):
        self.text = ''
        self._sentences = []
        self._entities = []
        self._ents = []

    @property
    def sentences(self):
        return self._sentences

    @sentences.setter
    def sentences(self, sentences):
        self._sentences = sentences

    def add_sentence(self, sentence):
        self._sentences.append(sentence)

    @property
    def entities(self):
        for sent in self._sentences:
            self._entities.append(sent.entities)
        return self._entities

    @property
    def ents(self):
        for sent in self._sentences:
            self._ents.append(sent.ents)
        return self._ents

    # def generate_html_file(self, output_file):
    #     return generate_html(self, output_file=output_file)

    # def annotate_text(self, output_file):
    #     annotated_sentences = [sent.annotated_text for sent in self._sentences]
    #     annotated_text = '\n'.join(annotated_sentences)
    #
    #     with open(output_file, 'w+') as fo:
    #         fo.write(annotated_text)
    #
    #     return annotated_text



from collections import Counter
import logging

from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)

# This function is modified from the function of stanza/models/ner/scorer.py
# This function to calculate the separate metrics of each entity.
#
# Copyright 2019 The Board of Trustees of The Leland Stanford Junior University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def score_by_entity(pred_tag_sequences, gold_tag_sequences, verbose=True):
    assert (len(gold_tag_sequences) == len(pred_tag_sequences)), \
        "Number of predicted tag sequences does not match gold sequences."

    def decode_all(tag_sequences):
        # decode from all sequences, each sequence with a unique id
        ents = []
        for sent_id, tags in enumerate(tag_sequences):
            for ent in decode_from_bio2(tags):
                ent['sent_id'] = sent_id
                ents += [ent]
        return ents

    gold_ents = decode_all(gold_tag_sequences)
    pred_ents = decode_all(pred_tag_sequences)

    correct_by_type = Counter()
    guessed_by_type = Counter()
    gold_by_type = Counter()

    # Added. Store all the entities
    entities = set()

    # records the details of fp and fn
    fp = []
    fn = []
    for p in pred_ents:
        if p not in gold_ents:
            fp.append(p)
    for p in gold_ents:
        if p not in pred_ents:
            fn.append(p)

    logger.info('Predict entities in total: {}'.format(len(pred_ents)))
    logger.info('Gold entities in total: {}'.format(len(gold_ents)))
    logger.info('False Positive: {}'.format(len(fp)))
    logger.info('False Negative: {}'.format(len(fn)))

    for p in pred_ents:
        guessed_by_type[p['type']] += 1
        if p in gold_ents:
            correct_by_type[p['type']] += 1
    for g in gold_ents:
        gold_by_type[g['type']] += 1
        entities.add(g['type'])

    prec_micro = 0.0
    if sum(guessed_by_type.values()) > 0:
        prec_micro = sum(correct_by_type.values()) * 1.0 / sum(guessed_by_type.values())
    rec_micro = 0.0
    if sum(gold_by_type.values()) > 0:
        rec_micro = sum(correct_by_type.values()) * 1.0 / sum(gold_by_type.values())
    f_micro = 0.0
    if prec_micro + rec_micro > 0:
        f_micro = 2.0 * prec_micro * rec_micro / (prec_micro + rec_micro)

    if verbose:
        logger.info("Prec.\tRec.\tF1")
        logger.info("{:.2f}\t{:.2f}\t{:.2f}".format(prec_micro * 100, rec_micro * 100, f_micro * 100))

    # metrics for entities
    if verbose:
        logger.info("Entity\tPrec.\tRec.\tF1")
        for entity in entities:
            prec_ent = 0.0
            if guessed_by_type[entity] > 0:
                prec_ent = correct_by_type[entity] * 1.0 / guessed_by_type[entity]
            rec_ent = 0.0
            if gold_by_type[entity] > 0:
                rec_ent = correct_by_type[entity] * 1.0 / gold_by_type[entity]
            f_ent = 0.0
            if prec_ent + rec_ent > 0:
                f_ent = 2.0 * prec_ent * rec_ent / (prec_ent + rec_ent)

            logger.info("{}\t{:.2f}\t{:.2f}\t{:.2f}".format(entity, prec_ent * 100, rec_ent * 100, f_ent * 100))

    return prec_micro, rec_micro, f_micro, fp, fn


# This function is modified from the function of stanza/models/ner/utils.py
# This function to decode all the entities in the sentence tags
#
# Copyright 2019 The Board of Trustees of The Leland Stanford Junior University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
def decode_from_bio2(tags):
    res = []
    ent_idxs = []
    cur_type = None

    def flush():
        if len(ent_idxs) > 0:
            res.append({
                'start': ent_idxs[0],
                'end': ent_idxs[-1],
                'type': cur_type})

    for idx, tag in enumerate(tags):
        if tag is None:
            tag = 'O'
        if tag == 'O':
            flush()
            ent_idxs.clear()
        elif tag.startswith('B-'):  # start of new ent
            flush()
            ent_idxs.clear()
            ent_idxs.append(idx)
            cur_type = tag[2:]
        elif tag.startswith('I-'):  # continue last ent
            ent_idxs.append(idx)
            cur_type = tag[2:]

    # flush after whole sentence
    flush()
    return res



# ---------------------------------------------------------------
# Train models using three NLP toolkits: spaCy, Stanza and FLAIR
#
# (C) 2020 Leibo Liu, Sydney NSW, Australia
# Released under Apache License 2.0
# ---------------------------------------------------------------
import logging
from pathlib import Path
from shutil import copyfile

from cdeid.data.data_loader import load_doc
from cdeid.models.ensemble_model import EnsembleModel
from cdeid.models.flair_trainer import FlairTrainer
from cdeid.models.spacy_trainer import SpacyTrainer
from cdeid.models.stanza_trainer import StanzaTrainer
from cdeid.data.data_operator import pass_bio_checking, sample_data_lines, parse_documents, load_data
from cdeid.utils.resources import PROGRESS_STATUS, PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


class Trainer:
    def __init__(self,
                 data_dir,
                 workspace,
                 phi_types,
                 wordvec_file,
                 resume_training=False,
                 train_file='train.bio',
                 dev_file='dev.bio',
                 test_file='test.bio'
                 ):
        self.data_dir = Path(data_dir)
        if not self.data_dir.exists():
            raise Exception('Data path does not exist: {}'.format(data_dir))

        self.workspace = workspace
        self.phi_types = phi_types
        self.wordvec_file = wordvec_file

        self.resume = resume_training

        self.train_file = train_file
        self.dev_file = dev_file
        self.test_file = test_file

    def train(self):
        """Train the models using the provided corpus

        Load the corpus of training and development and train models on balanced and
        imbalanced data sets.

        Args:

        Returns:
            no value
        """
        # prepare workspace
        self.prepare_workspace()

        # 0. check the status of training in the .status file in the workspace
        status_flags = [1] * 8

        if not self.resume:
            logger.info('A new training process is starting')
        else:
            status = self.check_train_status()
            if status - 1:
                for i in range(0, status - 1):
                    status_flags[i] = 0

            logger.info('Resume the previous training process at the step {} - {}'
                        .format(status, PROGRESS_STATUS[status]))

        # Process training
        if status_flags[0]:
            self.prepare_datasets()

        if status_flags[1]:
            self.train_spacy_balanced()

        if status_flags[2]:
            self.train_stanza_balanced()

        if status_flags[3]:
            self.train_flair_balanced()

        if status_flags[4]:
            self.train_spacy_imbalanced()

        if status_flags[5]:
            self.train_stanza_imbalanced()

        if status_flags[6]:
            self.train_flair_imbalanced()

        if status_flags[7]:
            self.ensemble_models()

    # check train status. Status stored in .status file in workspace with a status number
    def check_train_status(self):
        status_file = open(Path(self.workspace) / '.status', 'r')
        status = status_file.readline()
        status_file.close()
        if len(status.strip()) == 0:
            return 1
        return int(status)

    # 1-prepare data sets
    # 2-train spacy on balanced sets
    # 3-train stanza on balanced sets
    # 4-train flair on balanced sets
    # 5-train spacy on imbalanced sets
    # 6-train stanza on imbalanced sets
    # 7-train flair on imbalanced sets
    # 8-ensemble model
    def update_train_status(self, status):
        status_file = open(Path(self.workspace) / '.status', 'w')
        status_file.write(str(status))
        status_file.close()

    # 0-prepare workspace directory
    def prepare_workspace(self):
        workspace_dir = Path(self.workspace)
        if not workspace_dir.exists():
            workspace_dir.mkdir()

        models_dir = Path(self.workspace) / 'models'
        if not models_dir.exists():
            models_dir.mkdir()

        data_dir = Path(self.workspace) / 'data'
        if not data_dir.exists():
            data_dir.mkdir()

        status_file = Path(self.workspace) / '.status'
        if not status_file.is_file():
            status_file.touch()

        logger.info('workspace prepared: models and data directories are ready')

    # 1-prepare balanced and imbalanced training and dev data sets
    def prepare_datasets(self):
        self.update_train_status(1)
        logger.info('Training step 1 - prepare data sets - started')
        data_dir = Path(self.workspace) / 'data'
        # if not data_dir.exists():
        #     data_dir.mkdir()
        # copy data to workspace/data directory
        copyfile(self.data_dir / self.train_file, data_dir / self.train_file)
        copyfile(self.data_dir / self.dev_file, data_dir / self.dev_file)
        copyfile(self.data_dir / self.test_file, data_dir / self.test_file)
        # load data from files
        train = load_data(data_dir / self.train_file)
        if not pass_bio_checking(train):
            raise Exception('Please check the BIO format in the file {}'.format(self.train_file))

        dev = load_data(data_dir / self.dev_file)
        if not pass_bio_checking(dev):
            raise Exception('Please check the BIO format in the file {}'.format(self.dev_file))

        # parse data to list of lists for sampling the balanced data sets
        train_doc = parse_documents(train)
        logger.info('Training set: {} lines'.format(sum([len(doc) for doc in train_doc])))
        # generate balanced data sets
        train_balanced_size, _ = sample_data_lines(train_doc, 1.0, data_dir / 'balanced', self.train_file)
        logger.info('Training balanced set: {} lines'.format(train_balanced_size))

        dev_doc = parse_documents(dev)
        logger.info('Dev set: {} lines'.format(sum([len(doc) for doc in dev_doc])))
        dev_balanced_size, _ = sample_data_lines(dev_doc, 1.0, data_dir / 'balanced', self.dev_file)
        logger.info('Dev balanced set: {} lines'.format(dev_balanced_size))

        copyfile(self.data_dir / self.test_file, data_dir / 'balanced' / self.test_file)

        logger.info('Training step 1 - prepare data sets - completed')

    # 2-train spaCy model on balanced training and dev data sets
    def train_spacy_balanced(self):
        self.update_train_status(2)
        logger.info('Training step 2 - train spacy on balanced sets - started')
        spacy_trainer = SpacyTrainer(self.workspace, self.phi_types, self.train_file, self.dev_file)
        spacy_trainer.train()
        logger.info('Training step 2 - train spacy on balanced sets - completed')

    # 3-train Stanza model on balanced training and dev data sets
    def train_stanza_balanced(self):
        self.update_train_status(3)
        logger.info('Training step 3 - train stanza on balanced sets - started')
        stanza_trainer = StanzaTrainer(self.workspace, self.train_file, self.dev_file, self.wordvec_file)
        stanza_trainer.train()
        logger.info('Training step 3 - train stanza on balanced sets - completed')

    # 4-train FLAIR model on balanced training and dev data sets
    def train_flair_balanced(self):
        self.update_train_status(4)
        logger.info('Training step 4 - train flair on balanced sets - started')
        flair_trainer = FlairTrainer(self.workspace, self.train_file, self.dev_file, self.test_file)
        flair_trainer.train()
        logger.info('Training step 4 - train flair on balanced sets - completed')

    # 5-train spaCy model on imbalanced training and dev data sets
    def train_spacy_imbalanced(self):
        self.update_train_status(5)
        logger.info('Training step 5 - train spacy on imbalanced sets - started')
        spacy_trainer = SpacyTrainer(self.workspace, self.phi_types, self.train_file, self.dev_file, balanced=False)
        spacy_trainer.train()
        logger.info('Training step 5 - train spacy on imbalanced sets - completed')

    # 6-train Stanza model on imbalanced training and dev data sets
    def train_stanza_imbalanced(self):
        self.update_train_status(6)
        logger.info('Training step 6 - train stanza on imbalanced sets - started')
        stanza_trainer = StanzaTrainer(self.workspace,
                                       self.train_file, self.dev_file, self.wordvec_file, balanced=False)
        stanza_trainer.train()
        logger.info('Training step 6 - train stanza on imbalanced sets - completed')

    # 7-train FLAIR model on imbalanced training and dev data sets
    def train_flair_imbalanced(self):
        self.update_train_status(7)
        logger.info('Training step 7 - train flair on imbalanced sets - started')
        flair_trainer = FlairTrainer(self.workspace, self.train_file, self.dev_file, self.test_file, balanced=False)
        flair_trainer.train()
        logger.info('Training step 7 - train flair on imbalanced sets - completed')

    # 8-ensemble models
    def ensemble_models(self):
        self.update_train_status(8)
        logger.info('Training step 8 - ensemble model - started')
        logger.info('Evaluate the ensemble model on the test set')
        model_dir = Path(self.workspace) / 'models'
        test_set_path = Path(self.workspace) / 'data' / self.test_file
        ensemble_model = EnsembleModel(str(model_dir / 'balanced' / 'best-model.pt'),
                                       str(model_dir / 'balanced' / 'model-best'),
                                       str(model_dir / 'balanced' / 'stanza_model.pt'),
                                       str(model_dir / 'best-model.pt'),
                                       str(model_dir / 'model-best'),
                                       str(model_dir / 'stanza_model.pt')
                                       )
        doc_text, gold_tags = load_doc(test_set_path)
        ensemble_model.evaluate(doc_text, gold_tags)
        # ensemble_model.evaluate_with_batch(doc_text, gold_tags, batch_size=64)

        logger.info('Training step 8 - ensemble model - completed')
        # mark the training process as completed
        self.update_train_status(9)



import argparse
import sys

from cdeid.deidentifier.phi_deid import PHIDeid
from cdeid.models.trainer import Trainer
from cdeid.utils.resources import PACKAGE_NAME


def get_options(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog=PACKAGE_NAME)
    parser.add_argument('--command', type=str,
                        choices=[
                            'train',
                            'deid'], required=True)
    parser.add_argument('--data_dir', type=str, help='data sets directory')
    parser.add_argument('--train_file', type=str, default='train.bio', help='the file name of training set')
    parser.add_argument('--dev_file', type=str, default='dev.bio', help='the file name of development set')
    parser.add_argument('--test_file', type=str, default='test.bio', help='the file name of test set')
    parser.add_argument('--workspace', type=str, required=True,
                        help='the workplace which is used to store data and trained models')
    parser.add_argument('--resume_training', type=bool, default=False, help='resume the last training process')
    parser.add_argument('--phi_types', type=str, nargs='+', help='customized PHI types')

    parser.add_argument('--wordvec_file', type=str, help='wordvec files')

    #
    parser.add_argument('--deid_file', type=str, help='the file name to be de-identified')
    parser.add_argument('--deid_output_dir', type=str, help='the file directory to be de-identified')

    options = parser.parse_args(args)
    return options


def main():
    options = get_options()
    command = options.command

    if command == 'train':
        trainer = Trainer(options.data_dir,
                          options.workspace,
                          options.phi_types,
                          options.wordvec_file,
                          options.resume_training,
                          options.train_file,
                          options.dev_file,
                          options.test_file)
        trainer.train()
    elif command == 'deid':
        deider = PHIDeid(options.workspace,
                         options.deid_output_dir)
        doc = deider(options.deid_file)
        deider.output(doc)


if __name__ == '__main__':
    main()



def annotate_sent(sent):
    tmp_text = sent.text
    offset = 0
    for entity in sent.entities:
        tmp_text = tmp_text[:entity[0] + offset] + '<**' + entity[2] + '**>' + tmp_text[entity[1] + offset:]
        offset += (len(entity[2]) + 6 - (entity[1] - entity[0]))
    return tmp_text


def annotate_doc(doc, output_file):
    annotated_sentences = [annotate_sent(sent) for sent in doc.sentences]
    annotated_text = '\n'.join(annotated_sentences)
    with open(output_file, 'w+') as fo:
        fo.write(annotated_text)



import logging.config

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            }
        },
        "handlers": {
            "console": {
                "level": "INFO",
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "cdeid": {
                "level": "INFO",
                "handlers": ["console"],
                "propagate": False
            }
        }
    }
)



import spacy
from cdeid.utils.resources import *


def ent_tuple_to_bio(sentences, bio_scheme=BIO_SCHEME_2):
    nlp = spacy.load('en_core_web_sm')
    tokens_list = []
    tokens_bio_list = []
    for record_txt, record_entities in sentences:
        entities = record_entities.get('entities', '')
        entity_strs = []
        for start, end, entity_tag in entities:
            entity_strs.append((record_txt[start:end], entity_tag))

        # replace the words with tags
        for entity_str in entity_strs:
            record_txt = record_txt.replace(entity_str[0],
                                            TAG_BEGIN + entity_str[1] + " " + entity_str[0] + ' ' + TAG_END)

        # remove the \r\n at the end of text.
        record_txt = record_txt.replace('\n', '')

        doc = nlp(record_txt)
        tokens = []
        tokens_bio = []
        entity_begin_flag = False
        entity_middle_flag = False
        entity_type = ''
        entity_token_index = 0
        same_type_entity_together = False
        previous_entity_type = ''

        for token in doc:
            if len(token.text.strip()) == 0:
                continue
            if token.text.startswith(TAG_BEGIN):
                entity_begin_flag = True
                entity_type = token.text[11:]
                if previous_entity_type != '' and previous_entity_type == entity_type:
                    same_type_entity_together = True
                continue

            if token.text.startswith(TAG_END):
                entity_begin_flag = False
                entity_middle_flag = False
                entity_type = ''
                if entity_token_index == 1 and bio_scheme == BIO_SCHEME_1:
                    tokens_bio[len(tokens_bio) - 1] = tokens_bio[len(tokens_bio) - 1].replace("B-", "I-", 1)
                entity_token_index = 0
                continue

            if entity_begin_flag and (not entity_middle_flag):
                # tokens.append(token.text)
                entity_token_index += 1
                if (same_type_entity_together and bio_scheme == BIO_SCHEME_1) or bio_scheme == BIO_SCHEME_2:
                    tokens_bio.append('B-' + entity_type)
                else:
                    tokens_bio.append('I-' + entity_type)

                entity_middle_flag = True
                previous_entity_type = entity_type

            elif entity_begin_flag and entity_middle_flag:
                # tokens.append(token.text)
                tokens_bio.append('I-' + entity_type)
                entity_token_index += 1

            elif not entity_begin_flag:
                # tokens.append(token.text)
                previous_entity_type = ''
                same_type_entity_together = False
                tokens_bio.append('O')

            tokens.append(token.text)

        # print(tokens)
        # print(tokens_bio)
        if len(tokens) != len(tokens_bio):
            raise Exception('Token length and BIO tag length are not the same.')

        # append into list
        tokens_list.append(tokens)
        tokens_bio_list.append(tokens_bio)

    return tokens_list, tokens_bio_list


# convert BIOES with list of list data structure to BIO2
def bioes_to_bio2(doc_bioes_tags):
    new_doc_tags = []
    for sent_tags in doc_bioes_tags:
        new_sent_tags = []
        for tag in sent_tags:
            if tag == 'O':
                new_sent_tags.append(tag)
            else:
                if len(tag) < 2:
                    raise Exception('Invalid BIOES tag found:{}'.format(tag))
                else:
                    if tag[:2] == 'E-':
                        new_sent_tags.append('I-' + tag[2:])
                    elif tag[:2] == 'S-':
                        new_sent_tags.append('B-' + tag[2:])
                    else:
                        new_sent_tags.append(tag)
        new_doc_tags.append((new_sent_tags))

    return new_doc_tags


def biolu_to_bio2(doc_biolu_tags):
    new_doc_tags = []
    for sent_tags in doc_biolu_tags:
        new_sent_tags = []
        for tag in sent_tags:
            if tag == 'O':
                new_sent_tags.append(tag)
            else:
                if len(tag) < 2:
                    raise Exception('Invalid BIOLU tag found:{}'.format(tag))
                else:
                    if tag[:2] == 'L-':
                        new_sent_tags.append('I-' + tag[2:])
                    elif tag[:2] == 'U-':
                        new_sent_tags.append('B-' + tag[2:])
                    else:
                        new_sent_tags.append(tag)
        new_doc_tags.append((new_sent_tags))

    return new_doc_tags


def to_bio2(sentence_bio_tags):
    new_sent_tags = []
    for tag in sentence_bio_tags:
        if tag == 'O':
            new_sent_tags.append(tag)
        else:
            if len(tag) < 2:
                raise Exception('Invalid BIO tag found:{}'.format(tag))
            else:
                if tag[:2] == 'L-' or tag[:2] == 'E-':
                    new_sent_tags.append('I-' + tag[2:])
                elif tag[:2] == 'U-' or tag[:2] == 'S-':
                    new_sent_tags.append('B-' + tag[2:])
                else:
                    new_sent_tags.append(tag)

    return new_sent_tags


# strip the start/end whitespaces of entities
def entity_strip(token_list, tag_list):
    bio_tag_list = tag_list
    for i in range(len(bio_tag_list) - 2):
        if bio_tag_list[i] == 'O' \
                and bio_tag_list[i + 1].startswith('B-') \
                and bio_tag_list[i + 2].startswith('I-'):
            if token_list[i + 1].isspace():
                bio_tag_list[i + 2] = bio_tag_list[i + 1]
                bio_tag_list[i + 1] = 'O'
        if bio_tag_list[i].startswith('I-') \
                and bio_tag_list[i + 1].startswith('I-') \
                and bio_tag_list[i + 2] == 'O':
            if token_list[i + 1].isspace():
                bio_tag_list[i + 1] = 'O'
    return bio_tag_list



# This file is modified from the function of stanza/models/ner_tagger.py
# This file to help training stanza model and be called from stanza_trainer.py
#
# Copyright 2019 The Board of Trustees of The Leland Stanford Junior University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Entry point for training and evaluating an NER tagger.

This tagger uses BiLSTM layers with character and word-level representations, and a CRF decoding layer 
to produce NER predictions.
For details please refer to paper: https://nlp.stanford.edu/pubs/qi2018universal.pdf.
"""

import sys
import os
import time
from datetime import datetime
import argparse
import logging
import numpy as np
import random
import json
import torch
from torch import nn, optim

from stanza.models.ner.data import DataLoader
from stanza.models.ner.trainer import Trainer
from stanza.models.ner import scorer
from stanza.models.common import utils
from stanza.models.common.pretrain import Pretrain
from stanza.utils.conll import CoNLL
from stanza.models.common.doc import *
from stanza.models import _training_logging

from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


def parse_args(arguments):
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='data/ner', help='Root dir for saving models.')
    parser.add_argument('--wordvec_dir', type=str, default='extern_data/word2vec', help='Directory of word vectors')
    parser.add_argument('--wordvec_file', type=str, default='', help='File that contains word vectors')
    parser.add_argument('--train_file', type=str, default=None, help='Input file for data loader.')
    parser.add_argument('--eval_file', type=str, default=None, help='Input file for data loader.')

    parser.add_argument('--mode', default='train', choices=['train', 'predict'])
    parser.add_argument('--lang', type=str, help='Language')
    parser.add_argument('--shorthand', type=str, help="Treebank shorthand")

    parser.add_argument('--hidden_dim', type=int, default=256)
    parser.add_argument('--char_hidden_dim', type=int, default=100)
    parser.add_argument('--word_emb_dim', type=int, default=100)
    parser.add_argument('--char_emb_dim', type=int, default=100)
    parser.add_argument('--num_layers', type=int, default=1)
    parser.add_argument('--char_num_layers', type=int, default=1)
    parser.add_argument('--pretrain_max_vocab', type=int, default=100000)
    parser.add_argument('--word_dropout', type=float, default=0)
    parser.add_argument('--locked_dropout', type=float, default=0.0)
    parser.add_argument('--dropout', type=float, default=0.5)
    parser.add_argument('--rec_dropout', type=float, default=0, help="Word recurrent dropout")
    parser.add_argument('--char_rec_dropout', type=float, default=0, help="Character recurrent dropout")
    parser.add_argument('--char_dropout', type=float, default=0, help="Character-level language model dropout")
    parser.add_argument('--no_char', dest='char', action='store_false', help="Turn off character model.")
    parser.add_argument('--charlm', action='store_true',
                        help="Turn on contextualized char embedding using character-level language model.")
    parser.add_argument('--charlm_save_dir', type=str, default='saved_models/charlm',
                        help="Root dir for pretrained character-level language model.")
    parser.add_argument('--charlm_shorthand', type=str, default=None,
                        help="Shorthand for character-level language model training corpus.")
    parser.add_argument('--char_lowercase', dest='char_lowercase', action='store_true',
                        help="Use lowercased characters in charater model.")
    parser.add_argument('--no_lowercase', dest='lowercase', action='store_false', help="Use cased word vectors.")
    parser.add_argument('--no_emb_finetune', dest='emb_finetune', action='store_false',
                        help="Turn off finetuning of the embedding matrix.")
    parser.add_argument('--no_input_transform', dest='input_transform', action='store_false',
                        help="Do not use input transformation layer before tagger lstm.")
    parser.add_argument('--scheme', type=str, default='bioes', help="The tagging scheme to use: bio or bioes.")

    parser.add_argument('--sample_train', type=float, default=1.0, help='Subsample training data.')
    parser.add_argument('--optim', type=str, default='sgd', help='sgd, adagrad, adam or adamax.')
    parser.add_argument('--lr', type=float, default=0.1, help='Learning rate.')
    parser.add_argument('--min_lr', type=float, default=1e-4, help='Minimum learning rate to stop training.')
    parser.add_argument('--momentum', type=float, default=0, help='Momentum for SGD.')
    parser.add_argument('--lr_decay', type=float, default=0.5, help="LR decay rate.")
    parser.add_argument('--patience', type=int, default=3, help="Patience for LR decay.")

    parser.add_argument('--max_steps', type=int, default=200000)
    parser.add_argument('--eval_interval', type=int, default=500)
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--max_grad_norm', type=float, default=5.0, help='Gradient clipping.')
    parser.add_argument('--log_step', type=int, default=20, help='Print log every k steps.')
    parser.add_argument('--save_dir', type=str, default='saved_models/ner', help='Root dir for saving models.')
    parser.add_argument('--save_name', type=str, default=None, help="File name to save the model")

    parser.add_argument('--seed', type=int, default=1234)
    parser.add_argument('--cuda', type=bool, default=torch.cuda.is_available())
    parser.add_argument('--cpu', action='store_true', help='Ignore CUDA.')
    args = parser.parse_args(arguments)
    return args


def main(arguments):
    args = parse_args(arguments)

    torch.manual_seed(args.seed)
    np.random.seed(args.seed)
    random.seed(args.seed)
    if args.cpu:
        args.cuda = False
    elif args.cuda:
        torch.cuda.manual_seed(args.seed)

    args = vars(args)
    logger.info("Running tagger in {} mode".format(args['mode']))

    if args['mode'] == 'train':
        train(args)
    else:
        evaluate(args)


def train(args):
    utils.ensure_dir(args['save_dir'])
    model_file = args['save_dir'] + '/' + args['save_name'] if args['save_name'] is not None \
        else '{}/{}_nertagger.pt'.format(args['save_dir'], args['shorthand'])

    # load pretrained vectors
    if len(args['wordvec_file']) == 0:
        vec_file = utils.get_wordvec_file(args['wordvec_dir'], args['shorthand'])
    else:
        vec_file = args['wordvec_file']
    # do not save pretrained embeddings individually
    pretrain = Pretrain(None, vec_file, args['pretrain_max_vocab'], save_to_file=False)

    if args['charlm']:
        if args['charlm_shorthand'] is None:
            logger.info("CharLM Shorthand is required for loading pretrained CharLM model...")
            sys.exit(0)
        logger.info('Use pretrained contextualized char embedding')
        args['charlm_forward_file'] = '{}/{}_forward_charlm.pt'.format(args['charlm_save_dir'],
                                                                       args['charlm_shorthand'])
        args['charlm_backward_file'] = '{}/{}_backward_charlm.pt'.format(args['charlm_save_dir'],
                                                                         args['charlm_shorthand'])

    # load data
    logger.info("Loading data with batch size {}...".format(args['batch_size']))
    train_doc = Document(json.load(open(args['train_file'])))
    train_batch = DataLoader(train_doc, args['batch_size'], args, pretrain, evaluation=False)
    vocab = train_batch.vocab
    dev_doc = Document(json.load(open(args['eval_file'])))
    dev_batch = DataLoader(dev_doc, args['batch_size'], args, pretrain, vocab=vocab, evaluation=True)
    dev_gold_tags = dev_batch.tags

    # skip training if the language does not have training or dev data
    if len(train_batch) == 0 or len(dev_batch) == 0:
        logger.info("Skip training because no data available...")
        sys.exit(0)

    logger.info("Training tagger...")
    trainer = Trainer(args=args, vocab=vocab, pretrain=pretrain, use_cuda=args['cuda'])
    logger.info(trainer.model)

    global_step = 0
    max_steps = args['max_steps']
    dev_score_history = []
    best_dev_preds = []
    current_lr = trainer.optimizer.param_groups[0]['lr']
    global_start_time = time.time()
    format_str = '{}: step {}/{}, loss = {:.6f} ({:.3f} sec/batch), lr: {:.6f}'

    # LR scheduling
    if args['lr_decay'] > 0:
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(trainer.optimizer, mode='max', factor=args['lr_decay'], \
                                                               patience=args['patience'], verbose=True,
                                                               min_lr=args['min_lr'])
    else:
        scheduler = None

    # start training
    train_loss = 0
    while True:
        should_stop = False
        for i, batch in enumerate(train_batch):
            start_time = time.time()
            global_step += 1
            loss = trainer.update(batch, eval=False)  # update step
            train_loss += loss
            if global_step % args['log_step'] == 0:
                duration = time.time() - start_time
                logger.info(format_str.format(datetime.now().strftime("%Y-%m-%d %H:%M:%S"), global_step, \
                                              max_steps, loss, duration, current_lr))

            if global_step % args['eval_interval'] == 0:
                # eval on dev
                logger.info("Evaluating on dev set...")
                dev_preds = []
                for batch in dev_batch:
                    preds = trainer.predict(batch)
                    dev_preds += preds
                _, _, dev_score = scorer.score_by_entity(dev_preds, dev_gold_tags)

                train_loss = train_loss / args['eval_interval']  # avg loss per batch
                logger.info(
                    "step {}: train_loss = {:.6f}, dev_score = {:.4f}".format(global_step, train_loss, dev_score))
                train_loss = 0

                # save best model
                if len(dev_score_history) == 0 or dev_score > max(dev_score_history):
                    trainer.save(model_file)
                    logger.info("New best model saved.")
                    best_dev_preds = dev_preds

                dev_score_history += [dev_score]
                logger.info("")

                # lr schedule
                if scheduler is not None:
                    scheduler.step(dev_score)

            # check stopping
            current_lr = trainer.optimizer.param_groups[0]['lr']
            if global_step >= args['max_steps'] or current_lr <= args['min_lr']:
                should_stop = True
                break

        if should_stop:
            break

        train_batch.reshuffle()

    logger.info("Training ended with {} steps.".format(global_step))

    best_f, best_eval = max(dev_score_history) * 100, np.argmax(dev_score_history) + 1
    logger.info("Best dev F1 = {:.2f}, at iteration = {}".format(best_f, best_eval * args['eval_interval']))


def evaluate(args):
    # file paths
    model_file = args['save_dir'] + '/' + args['save_name'] if args['save_name'] is not None \
        else '{}/{}_nertagger.pt'.format(args['save_dir'], args['shorthand'])

    # load model
    use_cuda = args['cuda'] and not args['cpu']
    trainer = Trainer(model_file=model_file, use_cuda=use_cuda)
    loaded_args, vocab = trainer.args, trainer.vocab

    # load config
    for k in args:
        if k.endswith('_dir') or k.endswith('_file') or k in ['shorthand', 'mode', 'scheme']:
            loaded_args[k] = args[k]

    # load data
    logger.info("Loading data with batch size {}...".format(args['batch_size']))
    doc = Document(json.load(open(args['eval_file'])))
    batch = DataLoader(doc, args['batch_size'], loaded_args, vocab=vocab, evaluation=True)

    logger.info("Start evaluation...")
    preds = []
    for i, b in enumerate(batch):
        preds += trainer.predict(b)

    gold_tags = batch.tags
    _, _, score = scorer.score_by_entity(preds, gold_tags)

    logger.info("NER tagger score:")
    logger.info("{} {:.2f}".format(args['shorthand'], score * 100))



from collections import Counter
import re
import spacy
import stanza
from flair.models import SequenceTagger as flair
from flair.data import Sentence as FlairSentence, build_spacy_tokenizer

from cdeid.data.data_loader import concatenate_sents
from cdeid.utils.converter import to_bio2, entity_strip
from cdeid.data.document import Document, Sentence, Token
import logging

from cdeid.utils.scorer import score_by_entity
from cdeid.utils.resources import SPACY_PRETRAINED_MODEL_LG, PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)
nlp = spacy.load(SPACY_PRETRAINED_MODEL_LG)


class EnsembleModel:
    def __init__(self, flair_model, spacy_model, stanza_model,
                 flair_model_imbalanced, spacy_model_imbalanced, stanza_model_imbalanced):
        # Stanza model object
        self.stanza_model = stanza.Pipeline(
            lang='en',
            processors={'tokenize': 'spacy'},
            ner_model_path=stanza_model,
            tokenize_no_ssplit=True
        )

        # Flair model object
        self.flair_model = flair.load(flair_model)

        # SpaCy model object
        self.spacy_model = spacy.load(spacy_model)

        # imbalanced
        # Stanza model object
        self.stanza_model_imbalanced = stanza.Pipeline(
            lang='en',
            processors={'tokenize': 'spacy'},
            ner_model_path=stanza_model_imbalanced,
            tokenize_no_ssplit=True
        )

        # Flair model object
        self.flair_model_imbalanced = flair.load(flair_model_imbalanced)

        # SpaCy model object
        self.spacy_model_imbalanced = spacy.load(spacy_model_imbalanced)

    def predict(self, text, remove_extra_whitespaces=False):
        # This flag will remove the extra whitespaces before predict
        if remove_extra_whitespaces:
            text = re.sub(' +', ' ', text)

        def majority_vote(tag1, tag2, tag3, tag4, tag5, tag6):
            # check the length
            if len(tag1) != len(tag2) or len(tag1) != len(tag3):
                raise Exception('The length of predictions different')
            # convert BIOLU and BIOES to BIO2 format
            stanza_preds_bio2 = to_bio2(tag1)
            spacy_preds_bio2 = to_bio2(tag2)

            stanza_preds_bio2_imbalanced = to_bio2(tag4)
            spacy_preds_bio2_imbalanced = to_bio2(tag5)

            final_tags = []
            tmp_tags = zip(tag3, spacy_preds_bio2, stanza_preds_bio2,
                             tag6, spacy_preds_bio2_imbalanced, stanza_preds_bio2_imbalanced)
            for token_tag in tmp_tags:
                tag = Counter(list(token_tag))
                common_tag = tag.most_common(1)[0][0]
                final_tags.append(common_tag)

            return final_tags

        def predict_line(line):
            logger.debug('Predict sentence {}'.format(line))
            sentence = Sentence()
            sentence.text = line
            if line.isspace() or line == '':
                return sentence, []
            # spacy predict
            logger.debug('Use Spacy')
            doc_spacy = self.spacy_model(line)
            preds_spacy = [token.ent_iob_
                           if token.ent_iob_ == 'O' else (token.ent_iob_ + '-' + token.ent_type_)
                           for token in doc_spacy]

            logger.debug('Use Stanza')
            doc_stanza = self.stanza_model.processors['tokenize'].process(line)
            doc_stanza = self.stanza_model.processors['ner'].process(doc_stanza)
            preds_stanza = [token.ner for sent in doc_stanza.sentences for token in sent.tokens]

            # flair predict
            flair_sentence = FlairSentence(line, use_tokenizer=build_spacy_tokenizer(self.spacy_model))
            # token list by Spacy Tokenizer
            tokens = [token.text for token in doc_spacy]

            logger.debug('Use Flair')
            self.flair_model.predict(flair_sentence)

            preds_flair = [token.get_tag('ner').value for token in flair_sentence]

            # imbalanced
            # spacy predict
            logger.debug('Use Spacy')
            doc_spacy_imbalanced = self.spacy_model_imbalanced(line)
            preds_spacy_imbalanced = [token.ent_iob_
                                      if token.ent_iob_ == 'O' else (token.ent_iob_ + '-' + token.ent_type_)
                                      for token in doc_spacy_imbalanced]

            logger.debug('Use Stanza')
            doc_stanza_imbalanced = self.stanza_model_imbalanced.processors['tokenize'].process(line)
            doc_stanza_imbalanced = self.stanza_model_imbalanced.processors['ner'].process(doc_stanza_imbalanced)
            preds_stanza_imbalanced = [token.ner for sent in doc_stanza_imbalanced.sentences for token in sent.tokens]

            # flair predict
            flair_sentence_imbalanced = FlairSentence(line,
                                                      use_tokenizer=build_spacy_tokenizer(self.spacy_model))
            # token list by Spacy Tokenizer
            # tokens_imbalanced = [token.text for token in doc_spacy_imbalanced]

            logger.debug('Use Flair')
            self.flair_model_imbalanced.predict(flair_sentence_imbalanced)

            preds_flair_imbalanced = [token.get_tag('ner').value for token in flair_sentence_imbalanced]

            # ner_tag = majority_vote(preds_spacy, preds_stanza, preds_flair)
            ner_tag = majority_vote(preds_stanza, preds_spacy, preds_flair,
                                    preds_stanza_imbalanced, preds_spacy_imbalanced, preds_flair_imbalanced)
            logger.debug('Get final tags')

            # remove whitespaces from begin and end of the entity
            if not remove_extra_whitespaces:
                ner_tag = entity_strip(tokens, ner_tag)

            # add token object into sentence
            for i in range(len(tokens)):
                token = Token(tokens[i], ner_tag[i])
                sentence.add_token(token)

            return sentence, list(zip(tokens, ner_tag))

        # split newlines to a list of text
        text_list = text.splitlines()
        # List of list of tuple (token, bio_tag). e.g. [[(token, bio_tag), (token, bio_tag)]]
        ner_preds = []
        # Document object to store the data
        doc = Document()
        doc.text = text
        logger.info('Start Predicting. {} lines'.format(len(text_list)))
        # preds_result = list(map(predict_line, text_list))
        preds_result = [predict_line(line) for line in text_list]
        logger.info('Complete Predicting.')
        for result in preds_result:
            doc.add_sentence(result[0])
            ner_preds.append(result[1])

        return doc, ner_preds

    def evaluate(self, text, gold_tags):
        doc_text = concatenate_sents(text)
        _, pred_tags = self.predict(doc_text)
        # pred_tags include token text. Need to transfer to bio only
        pred_ner = [[token[1] for token in sent] for sent in pred_tags]
        _, _, _, fp, fn = score_by_entity(pred_ner, gold_tags)
        # print fp and fn
        logger.info('-------------False Positive Entities---------------')
        logger.info('Entity Number: {}'.format(len(fp)))
        for fp_ent in fp:
            sent_text = text[fp_ent['sent_id']]
            logger.info('Entities: {} @ {}'.format(fp_ent, sent_text))
            # print('Entities: {} @ {}'.format(fp_ent, sent_text))
        logger.info('-------------False Negative Entities---------------')
        logger.info('Entity Number: {}'.format(len(fn)))
        for fn_ent in fn:
            sent_text = text[fn_ent['sent_id']]
            logger.info('Entities: {} @ {}'.format(fn_ent, sent_text))



from mako.template import Template
from pathlib import Path

HTML_TAG_1 = '<span class="entity '
HTML_TAG_2 = '"><span class="words">'
HTML_TAG_3 = '</span><span class="ner">'
HTML_TAG_4 = '</span></span>'


def add_html_tags_of_entities(text, entities):
    offset = 0
    for entity in entities:
        tag_before = HTML_TAG_1 + entity[2].lower() + HTML_TAG_2
        tag_after = HTML_TAG_3 + entity[2].upper() + HTML_TAG_4

        text = text[:(entity[0] + offset)] \
               + tag_before + text[(entity[0] + offset):(entity[1] + offset)] \
               + tag_after + text[(entity[1] + offset):]

        offset += (len(tag_before) + len(tag_after))

    return text


def generate_html(doc, output_file='./display.html', template='display_template.html'):
    template_file = str(Path(__file__).parent.absolute() / template)
    temp = Template(filename=template_file)
    tagged_lines = []
    for sent in doc.sentences:
        tagged_lines.append(add_html_tags_of_entities(sent.text, sent.entities))

    html_content = temp.render(doc=tagged_lines)

    with open(output_file, 'w+') as fo:
        fo.write(html_content)



PACKAGE_NAME = 'cdeid'
SPACY_PRETRAINED_MODEL_LG = 'en_core_web_lg'
# SPACY_PRETRAINED_MODEL_SM = 'en_core_web_sm'
PROGRESS_STATUS = {
    1: 'prepare data sets',
    2: 'train spacy on balanced sets',
    3: 'train stanza on balanced sets',
    4: 'train flair on balanced sets',
    5: 'train spacy on imbalanced sets',
    6: 'train stanza on imbalanced sets',
    7: 'train flair on imbalanced sets',
    8: 'ensemble models',
    9: 'previous training process completed.'
}

TAG_BEGIN = 'TAGTAGBEGIN'
TAG_END = 'TAGTAGEND'
BIO_SCHEME_1 = 'BIO1'
BIO_SCHEME_2 = 'BIO2'

tag_1 = '<span class="selWords">'
tag_2 = '<span class='
tag_3 = 'tag">'
tag_4 = '</span></span>'


from stanza.utils.prepare_ner_data import load_conll03
from pathlib import Path
import json
import logging
import cdeid.models.stanza_trainer_agent as stanza_agent
from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


def bio_to_json(data_dir, file_name):
    sentences = load_conll03(Path(data_dir) / file_name)
    logger.info("{} examples loaded from {}".format(len(sentences), Path(data_dir) / file_name))

    document = []
    for (words, tags) in sentences:
        sent = []
        for w, t in zip(words, tags):
            sent += [{'text': w, 'ner': t}]
        document += [sent]

    output_file_name = 'stanza_' + str(Path(Path(file_name).parts[-1]).with_suffix('.json'))
    with open(Path(data_dir) / output_file_name, 'w') as outfile:
        json.dump(document, outfile)
    logger.info("Generated json file {}.".format(Path(data_dir) / output_file_name))
    return output_file_name


class StanzaTrainer:
    def __init__(self,
                 workspace,
                 train_file,
                 dev_file,
                 wordvec_file,
                 balanced=True):
        self.workspace = workspace
        self.train_file = train_file
        self.dev_file = dev_file
        self.wordvec_file = wordvec_file
        self.balanced = balanced

    def get_work_dir(self):
        if self.balanced:
            return 'balanced'
        return ''

    def convert_data(self):
        balanced = self.get_work_dir()
        data_dir = Path(self.workspace) / 'data' / balanced

        train_file_name = bio_to_json(data_dir, self.train_file)
        dev_file_name = bio_to_json(data_dir, self.dev_file)
        return train_file_name, dev_file_name

    def download_wordvec(self):
        # download and save wordvec in .cdeid directory in user home for the first time or importlib_resources package
        # wordvec_dir = Path.home() / '.' / PACKAGE_NAME
        # check if wordvec file already in the folder of word2vec
        # if not, download from a github link
        # if yes, return the full path of the en.vectors.xz
        return self.wordvec_file

    def train(self):
        logger.info('Convert training data')
        train_file_name, dev_file_name = self.convert_data()
        logger.info('Download pre-trained word2vec embeddings of CoNLL2017')
        wordvec_path = self.download_wordvec()
        data_dir = Path(self.workspace) / 'data' / self.get_work_dir()
        model_path = Path(self.workspace) / 'models' / self.get_work_dir()

        args = ['--train_file', str(data_dir / train_file_name), '--eval_file', str(data_dir / dev_file_name),
                '--lang', 'en', '--mode', 'train', '--wordvec_file', wordvec_path, '--save_dir', str(model_path),
                '--save_name', 'stanza_model.pt']

        stanza_agent.main(args)



from setuptools import setup, find_packages

with open('README.md', "r") as f:
    readme = f.read()

with open('LICENSE') as f:
    license_content = f.read()

setup(
    name='cdeid',
    version='0.1.2',
    author='Leibo Liu',
    author_email='liuleibo@gmail.com',
    description='A Customized De-identification framework',
    long_description_content_type='text/markdown',
    long_description=readme,
    url='https://github.com/CBDRH/cdeid',
    keywords=['DE-IDENTIFICATION', 'NLP'],
    install_requires=[
        'spaCy>=2.3.2',
        'stanza>=1.1.1',
        'flair==0.4.5',
        'mako>=1.1.3'
    ],
    packages=find_packages(exclude=('tests', 'docs')),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7'
)



import logging

from cdeid.deidentifier.annotator import annotate_doc
from cdeid.display.html_generator import generate_html
from cdeid.models.ensemble_model import EnsembleModel
from cdeid.utils.resources import PACKAGE_NAME
from pathlib import Path

logger = logging.getLogger(PACKAGE_NAME)


class PHIDeid:
    def __init__(self,
                 workspace,
                 deid_output_dir):
        logger.info('Loading models......')
        model_dir = Path(workspace) / 'models'
        self.model = EnsembleModel(str(model_dir / 'balanced' / 'best-model.pt'),
                                       str(model_dir / 'balanced' / 'model-best'),
                                       str(model_dir / 'balanced' / 'stanza_model.pt'),
                                       str(model_dir / 'best-model.pt'),
                                       str(model_dir / 'model-best'),
                                       str(model_dir / 'stanza_model.pt'))
        logger.info('Model loaded......')
        self.deid_output_dir = deid_output_dir
        # self.deid_file = deid_file

    def __call__(self, deid_file):
        self.deid_file = deid_file
        with open(self.deid_file, 'r') as f:
            content = f.read()
            doc, _ = self.model.predict(content)

        return doc

    def output(self, doc):
        html_file = self.deid_output_dir + '/' \
                    + str(Path(Path(self.deid_file).parts[-1]).with_suffix('.html'))
        generate_html(doc, html_file)

        annotated_file = self.deid_output_dir + '/' \
                         + str(Path(Path(self.deid_file).parts[-1]).with_suffix('.annotated.txt'))
        annotate_doc(doc, annotated_file)






import random
import logging
from pathlib import Path

from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


def load_data(data_file):
    data = open(data_file).readlines()
    return data


def pass_bio_checking(lines):
    for i, line in enumerate(lines):
        if ('-DOCSTART-' in line) or line.strip() == '':
            continue
        # the NER tag in the last column
        ner_tag = line.split()[-1]
        if not (ner_tag.startswith('B-') or ner_tag.startswith('I-') or ner_tag.startswith('O')):
            logger.error('Input file does not have correct BIO format at line {}.'.format(i))
            return False
    return True


def parse_documents(content):
    docs = []
    doc = []
    line = []
    # convert doc to a list of lines
    for item in content:
        if '-DOCSTART-' in item:
            if len(doc) != 0:
                docs.append(doc)
                doc = []
            continue
        if item == '' or item == '\n':
            if len(line) != 0:
                doc.append(line)
                line = []
            continue
        line.append(item)

    if len(doc) != 0:
        docs.append(doc)

    return docs


# proportion will be like 1.0 or 1.5 or 2.0 or 4.0.
# Text lines with PHI is 10. 1.0 means the lines without PHI will be 10 * 1.0
def sample_data_lines(docs, proportion, output_file, file_name):
    all_lines = [line for doc in docs for line in doc]

    lines_with_phi = [line for line in all_lines if sum('B-' in token for token in line) > 0]
    lines_without_phi = [line for line in all_lines if sum('B-' in token for token in line) == 0]

    logger.info('PHI lines: {}, No PHI lines: {}'.format(len(lines_with_phi), len(lines_without_phi)))

    random.Random(2020).shuffle(lines_without_phi)
    selected_lines_without_phi = random.sample(lines_without_phi, int(len(lines_with_phi) * proportion))

    all_lines = lines_with_phi + selected_lines_without_phi

    random.Random(2020).shuffle(all_lines)
    if not Path(output_file).exists():
        Path(output_file).mkdir()
    with open(Path(output_file) / file_name, 'w', newline='\n') as f:
        for line in all_lines:
            for token in line:
                f.write(token)
            f.write('\n')
        logger.info('Data file {} created.'.format(Path(output_file) / file_name))

    # sampled_doc = []
    # for line in all_lines:
    #     sampled_doc += line
    #     sampled_doc += ['\n']

    return len(all_lines), len(lines_with_phi)



from flair.data import Corpus
from flair.datasets import ColumnCorpus
from flair.embeddings import TokenEmbeddings, WordEmbeddings, StackedEmbeddings
from flair.models import SequenceTagger
from flair.trainers import ModelTrainer
from typing import List
from pathlib import Path


class FlairTrainer:
    def __init__(self,
                 workspace,
                 train_file,
                 dev_file,
                 test_file,
                 balanced=True):
        self.workspace = workspace
        self.train_file = train_file
        self.dev_file = dev_file
        self.test_file = test_file
        self.balanced = balanced

    def get_work_dir(self):
        if self.balanced:
            return 'balanced'
        return ''

    def train(self):
        # define the columns of data files
        columns = {0: 'text', 1: 'ner'}

        # data dir
        data_dir = Path(self.workspace) / 'data' / self.get_work_dir()
        model_path = Path(self.workspace) / 'models' / self.get_work_dir()

        corpus: Corpus = ColumnCorpus(data_dir, columns,
                                      train_file=self.train_file,
                                      dev_file=self.dev_file,
                                      test_file=self.test_file)

        tag_dictionary = corpus.make_tag_dictionary(tag_type='ner')
        embedding_types: List[TokenEmbeddings] = [
            WordEmbeddings('glove'),
        ]
        embeddings: StackedEmbeddings = StackedEmbeddings(embeddings=embedding_types)

        tagger: SequenceTagger = SequenceTagger(hidden_size=256,
                                                embeddings=embeddings,
                                                tag_dictionary=tag_dictionary,
                                                tag_type='ner',
                                                use_crf=True)

        trainer: ModelTrainer = ModelTrainer(tagger, corpus)

        trainer.train(model_path,
                      learning_rate=0.1,
                      mini_batch_size=32,
                      max_epochs=150)



import spacy
from spacy.cli import download, convert, train
from pathlib import Path
import logging

from cdeid.utils.resources import PACKAGE_NAME

logger = logging.getLogger(PACKAGE_NAME)


class SpacyTrainer:
    def __init__(self,
                 workspace,
                 phi_types,
                 train_file,
                 dev_file,
                 balanced=True):
        self.balanced = balanced
        self.workspace = workspace
        self.phi_types = phi_types
        self.train_file = train_file
        self.dev_file = dev_file

    def get_work_dir(self):
        if self.balanced:
            return 'balanced'
        return ''

    def convert_data(self):
        balanced = self.get_work_dir()
        convert(Path(self.workspace) / 'data' / balanced / self.train_file,
                Path(self.workspace) / 'data' / balanced,
                converter='ner')
        convert(Path(self.workspace) / 'data' / balanced / self.dev_file,
                Path(self.workspace) / 'data' / balanced,
                converter='ner')

    def customize_base_mode(self):
        if not Path(self.workspace + '/models/en_core_web_md').exists():
            # download en_core_web_lg model
            try:
                en_lg_model = spacy.load('en_core_web_lg')
            except:
                download('en_core_web_lg')
                en_lg_model = spacy.load('en_core_web_lg')

            # add customized NER types based on en_core_web_lg
            ner = en_lg_model.get_pipe('ner')
            for phi in self.phi_types:
                ner.add_label(phi)

            logger.info('Customized PHI Types added into the base model: {}'.format(str(self.phi_types)))
            en_lg_model.to_disk(Path(self.workspace) / 'models' / 'en_core_web_md')

        return self.workspace + '/models/en_core_web_md'

    def train(self):
        logger.info('Convert training data')
        self.convert_data()
        logger.info('Customize base model')
        base_model = self.customize_base_mode()
        data_dir = Path(self.workspace) / 'data' / self.get_work_dir()
        model_path = Path(self.workspace) / 'models' / self.get_work_dir()
        train_file = Path(Path(self.train_file).parts[-1]).with_suffix('.json')
        dev_file = Path(Path(self.dev_file).parts[-1]).with_suffix('.json')

        train('en', model_path, data_dir / train_file,
              data_dir / dev_file, base_model=base_model, pipeline='ner')
